#ifndef COMMON_H
#define COMMON_H

#define MAX(x,y) ((x)>(y)?(x):(y))
#define MIN(x,y) ((x)<(y)?(x):(y))
#define CLAMP(x, y, z) ((x)<(y)?(y):((x)>(z)?(z):(x)))

struct __textures {
	int count;
	int width[10], height[10], bpp[10];
	unsigned char *texture[10], *texture_device[10];
};

struct __dimensions {
	int width, height;
};

struct __offsets {
	int x, y;
};

struct __box {
	double minx, miny, minz;
	double maxx, maxy, maxz;
};

struct __ray {
	__vector origin;
	__vector direction;
};

struct __intersection {
	bool intersects;
	double t;
	
	__ray ray;
	__vector normal;
};

struct __object {
	int type; // geometry
	int material;
	int texture;
	__vector color;
	__vector emission;
	
	// for planes
	__vector normal;
	__vector point;
	__vector e0;
	__vector e1;

	// for spheres
	__vector center;
	double radius;
	
	// for triangles
	__vector p0, p1, p2;
	__vector n0, n1, n2;
	__vector n;
};


struct _bounding_box {
	unsigned short *depth,
		       *depth_device;
	double *minx, *miny, *minz, *maxx, *maxy, *maxz,
	       *minx_device, *miny_device, *minz_device, *maxx_device, *maxy_device, *maxz_device;
	short *child0, *child1,
	      *child0_device, *child1_device;
	short *parent,
	      *parent_device;

	short *id, *id_device;
	short *leaf_id, *leaf_id_device;

	unsigned short *n_objects,
		       *n_objects_device;
	unsigned short *objects,
		       *objects_device;
	unsigned short max_depth;
	unsigned short size, leaf_size;
};


struct _camera {
	double focal_length, aperture;
	double focal_distance, image_distance;
	double aperture_diameter;
};

bool initializePathTracer(_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device, __object* _objects, __object*& current_objects_device, int object_count, float*& rand_device, int max_bounces, __dimensions dim, __box bounds, __textures& texture);
bool setupPathTracer     (_bounding_box& b,                                                                              double*& current_doubles_device, __object* _objects, __object*& current_objects_device, int object_count,                                       __dimensions dim, __box bounds);
bool releasePathTracer   (_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device,                     __object*& current_objects_device,                   float*& rand_device,                                                  __textures& texture);
bool runPathTracer       (_bounding_box& b, unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, double*& current_doubles_device,                     __object*& current_objects_device, int object_count, float*& rand_device, int max_bounces, __dimensions dim,               __offsets offset, unsigned long long samples, _camera cmodel, __textures& texture);
bool grabFrame           (unsigned char*& current_buffer_host, unsigned char *& current_buffer_device, __dimensions dim);

#endif